import os.path

import fmover.cli as cli
from fmover.configs import MoveConfigsHandler
from fmover.mover import Mover
import appdirs
from fmover.base_logger import logger


def main():
    # Parse arguments
    args = cli.enable_argument_parser().parse_args()
    # Initialize Configuration Handler
    c_handler = MoveConfigsHandler(appdirs.user_config_dir("fmover"))

    # The user wants to list all configurations
    if args.listConfigs:
        c_handler.print_configs()

    # The user wants to open a given configuration file in a editor
    if args.openConfig:
        try:
            c_handler.open_config(args.openConfig)
        except (FileNotFoundError, OSError) as e:
            logger.error(e)
            exit(1)

    # The user wants to print the content of a given configuration file
    if args.printConfig:
        try:
            c_handler.print_config_content(args.printConfig)
        except FileNotFoundError as e:
            logger.error(e)
            exit(1)

    # The user wants to create a new configuration file with a given basename
    if args.createConfig:
        try:
            c_handler.create_config(args.createConfig)
            print("Configuration file created. The new list of configurations is:")
            c_handler.print_configs()
        except FileExistsError as e:
            input(f"Configuration file already exists. Do you want to overwrite it? [y/N] ")
            if input().lower() == "y":
                c_handler.delete_config(args.createConfig)
                c_handler.create_config(args.createConfig)

    # The user wants to delete a configuration file with a given basename
    if args.deleteConfig:
        try:
            answer = input(f"Are you sure to delete {args.deleteConfig}? (y/n)")
            while answer != "y" and answer != "n":
                answer = input("Please enter \"y\" or \"n\" ")
            if answer == "y":
                c_handler.delete_config(args.deleteConfig)
                print("Configuration file deleted. The new list of configurations is:")
                c_handler.print_configs()
        except FileNotFoundError as e:
            logger.error(e)
            exit(1)

    # The user wants to move a single file based on a given configuration and a given file path
    if args.moveOneFile:
        try:
            file_path, config = tuple(args.moveOneFile)
            if not os.path.isabs(file_path):
                file_path = os.path.abspath(file_path)
            Mover(c_handler.get_config_path(config)).move_file(file_path, not args.silent, args.force)
        except (ValueError, FileNotFoundError) as e:
            # The user did not provide a valid configuration file
            logger.error(e)
            exit(1)

    # The user wants to move all files in a directory based on a given configuration and a given directory path
    if args.moveAllFilesInFolder:
        try:
            dir_path, config = tuple(args.moveAllFilesInFolder)
            Mover(c_handler.get_config_path(config)).move_files_in_dir(dir_path, not args.silent, args.force)
        except (ValueError) as e:
            # The user did not provide a valid configuration file
            logger.error(e)
            exit(1)


if __name__ == '__main__':
    main()



import shutil
from os import listdir, getcwd, path, mkdir
from os.path import isfile, join, isabs, normpath, exists, dirname, isdir, relpath
import notifypy as notify
from fmover.config import MoveConfig
from fmover.file_property import FileMetadata
from fmover.interpreter import Interpreter
from fmover.base_logger import logger


class Mover:
    """
    This class is responsible for moving files to a new location based on a configuration file.

    Attributes:
        configuration (MoveConfig): A MoveConfig object that represents the configuration file

    Methods:
        move_file(file_path: str, should_notify: bool) -> None
            Moves a file to a new location based on the configuration file
        move_files_in_dir(path_to_directory: str, should_notify: bool) -> None
            Moves all files in a directory to a new location based on the configuration file
    """

    def __init__(self, move_config_path: str):
        self.configuration = MoveConfig(move_config_path)
        self.configuration.validate_config()

    def _get_new_file_location(self, file_data: FileMetadata) -> str:
        interpreter = Interpreter(file_properties=file_data.get_file_properties(),
                                  move_config=self.configuration.get_properties(),
                                  commands=self.configuration.get_commands())
        location = interpreter.parse_command()
        if location is None:
            location = dirname(file_data.file)
        return join(location, file_data.get_file_name_with_extension())

    def _notify(self, file_name, new_location) -> None:
        notification = notify.Notify()
        notification.application_name = "File has been moved"
        notification.title = file_name
        notification.message = new_location
        notification.send()

    def move_file(self, file_path: str, should_notify: bool, force: bool) -> None:
        """
        Moves a file to a new location based on the configuration file
        :param force: If true, the destination directory will be created if it does not exist
        :param file_path: The path of the file (can be a relative path)
        :param should_notify: If True, a pop-up notification will be shown
        """
        file_data = FileMetadata(file_path)
        logger.debug(f"File data: {file_data.get_file_properties()}")
        new_file_path = self._get_new_file_location(file_data)
        if not isabs(new_file_path):
            new_file_path = normpath(join(dirname(file_path), new_file_path))
        f_name = file_data.get_file_name_with_extension()
        dir_of_new_file = path.dirname(new_file_path)
        if file_path == new_file_path:
            logger.info(f"The file is already in the correct location: {f_name}")
            return

        if not exists(dir_of_new_file) and force:
            mkdir(dir_of_new_file)
        elif not exists(dir_of_new_file):
            logger.warning(f"The destination directory does not exist: {dir_of_new_file} "
                           f"\n if you wish to create directories if they do not exist, "
                           f"please add the -f flag")
            return
        if exists(new_file_path):
            logger.warning(f"A file with the same name already exists in the destination directory: {new_file_path}")
            return

        shutil.move(file_path, new_file_path)
        if should_notify:
            self._notify(file_data.get_file_name_with_extension(), new_file_path)
        logger.info(f"File successfully moved: {relpath(new_file_path, file_path)}")
        logger.debug(f"File successfully moved: {normpath(file_path)} -> {new_file_path}")

    def move_files_in_dir(self, path_to_directory: str, should_notify: bool, force: bool) -> None:
        """
        Moves all files in a directory to a new location based on the configuration file
        :param force: If true, the destination directory will be created if it does not exist
        :param path_to_directory: The path of the directory
        :param should_notify: If True, a pop-up notification will be shown
        """
        if not isabs(path_to_directory):
            path_to_directory = join(getcwd(), path_to_directory)
        if not isdir(path_to_directory):
            raise FileNotFoundError(f"The directory {path_to_directory} does not exist.")
        file_paths = [join(path_to_directory, file) for file in listdir(path_to_directory) if file[0] != "."
                      and isfile(join(path_to_directory, file))]
        for file_path in file_paths:
            self.move_file(file_path, should_notify, force)



from fmover.base_logger import logger


def _get_parameter(token: str) -> str:
    """
    :param token: a token of the form "PARAMETER(PATTERN)"
    :return: the parameter of the token
    """
    return token[0:token.find("(")].strip()


def _get_pattern(token: str) -> str:
    """
    :param token: a token of the form "PARAMETER(PATTERN)"
    :return: the pattern of the token
    """
    return token[token.find("(") + 1:token.find(")")].strip()


class Interpreter:
    """
    This class compares file properties with configuration properties based on commands.

    Example:
    file_properties = {"WHERE_FROM": "https://www.google.com", "NAME": "hello", "FILE_EXTENSION": ".txt"}
    move_config = {"WHERE_FROM": {"https://www.google.com": "C:/Users/Downloads/Google", "https://www.youtube.com": "C:/Users/Downloads/Youtube"},
                    "NAME": {"hello": "C:/Users/Downloads/Hello", "world": "C:/Users/Downloads/World"},
                    "FILE_EXTENSION": {".txt": "C:/Users/Downloads/Text", ".pdf": "C:/Users/Downloads/PDF"}}
    commands = [{"WHERE_FROM(https://www.google.com) & NAME(hello)": "NAME(*)"}, {"WHERE_FROM(https://www.youtube.com) & NAME(world)": "NAME(*)"}]

    The Interpreter class will return the path "C:/Users/Downloads/Hello" because the antecedent of the first command is true.

    Attributes:
        file_properties: A dictionary of file properties
        config: A dictionary of configuration properties
        commands: A list of commands

    Methods:
        parse_command: Returns the path of the first command whose antecedent is true
        antecedent_corresponds: Returns true if all tokens in an antecedent are true (which are connected by conjunction)
        token_corresponds: Returns true if the parameter of the given token has a pattern that is contained in the
                           file property of the parameter
        get_corresponding_path: Returns the path of the command whose antecedent is true
    """

    # Dictionary:
    # Antecedent: The left side of a command
    # Consequent: The right side of a command
    # Token: The part of an antecedent which is separated by conjunction
    # Parameter: The left side of a token e.g. the parameter of NAME(hello) is NAME
    # Pattern: The content inside the brackets of a token e.g. the pattern of NAME(hello) is hello
    # Property: A configuration property e.g. Name, File Extension, Where From
    # *: A wildcard which indicates all pattern:path pairs of a parameter

    def __init__(self, file_properties: dict, move_config: dict, commands: list):
        self.file_properties = file_properties
        self.config = move_config
        self.commands = commands

    def parse_command(self) -> str:
        """
        :return: The path of the first command whose antecedent is true
        """
        for command in self.commands:
            (antecedent, consequent), = command.items()
            if self.__antecedent_corresponds(antecedent):
                logger.debug(f"Executed command: {command}")
                return self.__get_corresponding_path(consequent)

    def __antecedent_corresponds(self, antecedent: str) -> bool:
        """
        :param antecedent: The antecedent of a command
        :return: True if all tokens in an antecedent are true (which are connected by conjunction)
        """
        return all(self.__token_corresponds(token.strip()) for token in antecedent.split("&"))

    # This method returns true if the parameter of the given token has a pattern that is contained in the
    # file property of the parameter
    def __token_corresponds(self, token: str) -> bool:
        """
        :param token: The token of an antecedent
        :return: True if the parameter of the given token has a pattern that
        is contained in the file property of the parameter
        """
        parameter, pattern = _get_parameter(token), _get_pattern(token)
        property_dic: dict = self.config.get(parameter)
        file_property: str = self.file_properties.get(parameter)
        if parameter not in self.file_properties:
            raise ValueError(f"The parameter and property {parameter} is not supported by the program")
        if pattern == "*":
            return any(property_pattern in file_property for property_pattern in property_dic)
        elif pattern in file_property:
            return True
        return False

    def __get_corresponding_path(self, token: str) -> str:
        """
        :param token: The consequent of a command
        :return: The path of the command whose antecedent is true
        """
        parameter, pattern = _get_parameter(token), _get_pattern(token)
        property_dic: dict = self.config.get(parameter)
        if parameter not in self.file_properties:
            raise ValueError(f"The parameter and property {parameter} is not supported by the program")
        if pattern == "*":
            for parameter_pattern in property_dic:
                if parameter_pattern in self.file_properties.get(parameter):
                    return property_dic.get(parameter_pattern)
        else:
            return property_dic.get(pattern)



import os
import subprocess
import sys


class FileMetadata:
    """
    This class is used to get the file properties of a file.

    Attributes:
        file (str): The path to the file

    Methods:
        get_file_name_with_extension(): Returns the file name with the extension and without the path
        get_file_extension(): Returns the file extension
        get_file_name(): Returns the file name without the extension and without the path
        get_where_from(): Returns the source where the file was obtained from. If nothing is found, it returns "UNKNOWN"
        get_file_properties(): Returns a dictionary of the file properties "WHERE_FROM", "NAME", "FILE_EXTENSION"
    """

    def __init__(self, file: str):
        self.file = file

    def get_file_name_with_extension(self):
        """
        :return: the file name with the extension and without the path
        """
        return os.path.basename(self.file)

    def get_file_extension(self) -> str:
        """
        :return: the file extension
        """
        return os.path.splitext(self.file)[1]

    def get_file_name(self) -> str:
        """
        :return: the file name without the extension and without the path
        """
        return self.get_file_name_with_extension().split(".")[0]

    def get_where_from(self) -> str:
        """
        :return: the source where the file was obtained from. If nothing is found, it returns "UNKNOWN"
        """
        # TODO: This does only work on macos. Make it work on windows and linux
        if sys.platform == "darwin":
            try:
                where_from = subprocess.check_output(["mdls", "-name", "kMDItemWhereFroms", self.file]).decode().split("=")[1].strip()
                return where_from if where_from != "(null)" else "UNKNOWN"
            except subprocess.CalledProcessError:
                raise FileNotFoundError(f"Could not read where the file {self.file} was obtained from")
        return "UNKNOWN"

    def get_file_properties(self) -> dict:
        """
        :return: a dictionary of the file properties "WHERE_FROM", "NAME", "FILE_EXTENSION"
        """
        return {"WHERE_FROM": self.get_where_from(),
                "NAME": self.get_file_name(),
                "FILE_EXTENSION": self.get_file_extension()}



import json
import os
import subprocess
import sys


class MoveConfigsHandler:
    """
    This class is responsible for monitoring the configurations directory, updating the configurations and reporting
    about them.

    Attributes:
        CONFIGS_DIR: The absolute path to the configurations directory

    Methods:
        get_config_path(config_name): Returns the absolute path to the configuration file
        list_configs(): Returns a list of all the configuration files in the configurations directory
        print_configs(): Prints all the configurations in the configurations directory
        print_config_content(config_name): Prints the content of the configuration file
        open_config(config_name): Opens the configuration file in the default text editor
        delete_config(config_name): Deletes the configuration file
        create_config(config_name): Creates a new configuration file
        get_default(): Returns the default configuration
    """

    def __init__(self, path_to_data_dir):
        self.CONFIGS_DIR = os.path.join(path_to_data_dir, "configurations")
        if not os.path.exists(path_to_data_dir):
            os.mkdir(path_to_data_dir)
            os.mkdir(self.CONFIGS_DIR)
            with open(os.path.join(self.CONFIGS_DIR, "default.json"), 'w') as f:
                json.dump(self.get_default(), f, indent=2)

    def get_config_path(self, config_name) -> str:
        """
        :param config_name: The name of the configuration file without the extension
        :return: The absolute path to the configuration file
        """
        return os.path.join(self.CONFIGS_DIR, config_name + ".json")

    def list_configs(self) -> list:
        """
        :return: A list of all the configuration files in the configurations directory
        """
        return [f[:-5] for f in os.listdir(self.CONFIGS_DIR) if f.endswith(".json")]

    def print_configs(self) -> None:
        """
        Prints all the configurations in the configurations directory
        """
        configs = [c.replace('.json', '') for c in self.list_configs()]
        print(*configs, sep="\n")

    def print_config_content(self, config_name) -> None:
        """
        Prints the content of the configuration file
        :param config_name: The name of the configuration file without the extension
        """
        config_path = self.get_config_path(config_name)
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"The configuration file {config_name} does not exist. "
                                    f"Choose one from the following: {self.list_configs()}")
        with open(config_path, 'r') as f:
            print(f.read())

    def open_config(self, config_name) -> None:
        """
        Opens the configuration file in the default text editor
        :param config_name: The name of the configuration file without the extension
        """
        config_path = self.get_config_path(config_name)
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"The configuration file {config_name} does not exist. "
                                    f"Choose one from the following: {self.list_configs()}")
        if sys.platform == "win32":
            os.startfile(config_path)
        elif sys.platform == "darwin":
            subprocess.call(('open', config_path))
        else:
            subprocess.call(('xdg-open', config_path))



    def delete_config(self, config_name) -> None:
        """
        Deletes the configuration file
        :param config_name: The name of the configuration file without the extension
        """
        config_path = self.get_config_path(config_name)
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"The configuration file {config_name} does not exist."
                                    f" Choose one from the following: {self.list_configs()}")
        os.remove(config_path)

    def create_config(self, config_name) -> None:
        """
        Creates a new configuration file
        :param config_name: The name of the configuration file without the extension
        """
        config_path = self.get_config_path(config_name)
        if os.path.exists(config_path):
            raise FileExistsError(f"The configuration file {config_name} already exists. "
                                  f"These are the existing configurations: {self.list_configs()}")
        with open(config_path, 'w') as f:
            json.dump(self.get_template(), f, indent=2)

    def get_default(self):
        return {"COMMAND": [{"FILE_EXTENSION(*)": "FILE_EXTENSION(*)"}], "WHERE_FROM": {}, "NAME": {},
                "FILE_EXTENSION": {
                    ".pdf": "./PDF", ".ps": "./PDF", ".jpg": "./Photos", ".jpeg": "./Photos", ".Jpeg": "./Photos",
                    ".gif": "./Photos",
                    ".HEIC": "./Photos", ".heic": "./Photos", ".JPG": "./Photos", ".jp2": "./Photos",
                    ".png": "./Photos",
                    ".svg": "./Photos", ".mov": "./Videos", ".MOV": "./Videos", ".mp4": "./Videos", ".MP4": "./Videos",
                    ".m4v": "./Videos", ".M4V": "./Videos", ".avi": "./Videos", ".AVI": "./Videos", ".docx": "./Text",
                    ".odt": "./Text", ".doc": "./Text", ".pages": "./Text", ".txt": "./Text", ".rtf": "./Text",
                    ".pptx": "./PowerPoint", ".m4a": "./Audio", ".wave": "./Audio", ".wav": "./Audio"}}

    def get_template(self):
        return {"COMMAND": [{}], "WHERE_FROM": {}, "NAME": {}, "FILE_EXTENSION": {}}




import json


class MoveConfig:
    """
    This class represents a configuration file.

    A move configuration is a JSON file which consists of a command and properties e.g. NAME, FILE_EXTENSION,
    WHERE_FROM.

    Attributes:
        config: The path of the configuration file or a dictionary which contains the configuration

    Methods:
        get_config: Returns the configuration file as a dictionary
        get_commands: Returns a list of commands
        get_properties: Returns a dictionary of properties
    """

    def __init__(self, config):
        if isinstance(config, str):
            with open(config, "r") as file:
                self.config = json.load(file)
        elif isinstance(config, dict):
            self.config = config
        else:
            raise ValueError(f"The config must be a path (string) or a dictionary")

    def get_commands(self) -> list:
        """
        :return: A list of dictionaries which contains the commands
        """
        return self.config.get("COMMAND")

    def get_properties(self) -> dict:
        """
        :return: A dictionary of dictionaries which contains the properties
        """
        return {key: value for key, value in self.config.items() if key != "COMMAND"}

    def validate_config(self) -> bool:
        """
        A valid config is a dict which must have a key "COMMAND" with a value which is a list of singleton dictionaries.
        These singleton dictionaries are commands which consist of antecedents (as keys) and consequents (as values).
        An antecedent consist of tokens which are seperated by "&". A consequent only has one token.
        Tokens consist of a parameter and a pattern. The tokens have the following form: "PARAMETER(PATTERN)".
        The parameter can be one of the following: "NAME", "FILE_EXTENSION", "WHERE_FROM".
        The pattern can be a string or a wildcard "*".
        If the parameter is declared in the command it must be a property in the config.

        This method raises a ValueError if the config is invalid. It shows where the error occurred
        :return: True if the configuration is valid, otherwise it raises a ValueError
        """

        if "COMMAND" not in self.config:
            raise ValueError(f"The config does not have a key \"COMMAND\"")

        if not isinstance(self.config["COMMAND"], list):
            raise ValueError(f"The value of the key \"COMMAND\" is not a list of dictionaries")

        for command in self.config["COMMAND"]:
            if not isinstance(command, dict):
                raise ValueError(f"The command {command} is not a dictionary")
            if len(command) != 1:
                raise ValueError(f"The command {command} is not a singleton dictionary")

            antecedent, consequent = list(command.items())[0]
            antecedent_tokens = antecedent.split("&")
            consequent_token = consequent

            for token in antecedent_tokens:
                self.__validate_token(token)
            self.__validate_token(consequent_token)
        return True

    def __validate_token(self, token: str):
        if "(" not in token or ")" not in token:
            raise ValueError(
                f"The token {token} is not a valid token. A token must have the form \"PARAMETER(PATTERN)\"")
        parameter, pattern = token.split("(")
        parameter = parameter.strip()
        pattern = pattern.replace(")", "").strip()
        if parameter not in self.config:
            raise ValueError(f"The parameter {parameter} is not declared in the config")
        if pattern != "*" and pattern not in self.config[parameter]:
            raise ValueError(f"The pattern {pattern} is not declared in the config for the parameter {parameter}")



import argparse


def enable_argument_parser() -> argparse.ArgumentParser:
    argument_parser = argparse.ArgumentParser(
        prog="mover",
        description="Move files based on given rules and file properties"
    )

    argument_parser.add_argument(
        "-l", "--list",
        action="store_true",
        dest="listConfigs",
        help="List all configurations"
    )

    argument_parser.add_argument(
        "-o", "--open",
        type=str,
        dest="openConfig",
        help="Open configuration file")

    argument_parser.add_argument(
        "-p", "--print",
        type=str,
        dest="printConfig",
        help="Print configuration")

    argument_parser.add_argument(
        "-n", "--new",
        type=str,
        metavar="configName",
        dest="createConfig",
        help="Create configuration with given name")

    argument_parser.add_argument(
        "-d", "--delete",
        type=str,
        metavar="configName",
        dest="deleteConfig",
        help="Delete a configuration"
    )

    argument_parser.add_argument(
        "-m", "--move",
        nargs=2,
        dest="moveOneFile",
        metavar=("pathToFile", "configName"),
        type=str,
        help="Move given file based on given configuration")

    argument_parser.add_argument(
        "-a", "--all",
        metavar=("pathToFolder", "configName"),
        nargs=2,
        dest="moveAllFilesInFolder",
        type=str,
        help="Move all files in the given folder based on given configuration")

    argument_parser.add_argument(
        "-s", "--silent",
        action="store_true",
        dest="silent",
        help="Do not notify that the file has been moved")

    # Optional force argument which creates a destination directory if it does not exist
    argument_parser.add_argument(
        "-f", "--force",
        action="store_true",
        dest="force",
        help="Create destination directory if it does not exist")

    return argument_parser



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fmover']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0', 'notifypy>=1.0.3.0,<2.0.0.0']

entry_points = \
{'console_scripts': ['fmover = fmover.__main__:main']}

setup_kwargs = {
    'name': 'fmover',
    'version': '0.0.4',
    'description': 'Move files based on file properties and given criteria',
    'long_description': 'None',
    'author': 'Ephraim Siegfried',
    'author_email': 'ephraim.siegfried@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/EphraimSiegfried/fmover',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import appdirs
import logging
import os
import sys

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Create log file if it does not exist
# Paste python -c "import appdirs; print(appdirs.user_log_dir('fmover'))" in terminal to get the log directory
log_path = appdirs.user_log_dir(''.join([appdirs.user_log_dir("fmover"), "/fmover.log"]))
if not os.path.exists(appdirs.user_log_dir("fmover")):
    os.mkdir(appdirs.user_log_dir("fmover"))

# Create handlers
file_handler = logging.FileHandler(log_path)
console_handler = logging.StreamHandler(sys.stdout)
file_handler.setLevel(logging.DEBUG)
console_handler.setLevel(logging.INFO)

# Create formatters and add it to handlers
file_format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
console_format = logging.Formatter("%(message)s")
file_handler.setFormatter(file_format)
console_handler.setFormatter(console_format)

# Add handlers to the logger
logger.addHandler(file_handler)
logger.addHandler(console_handler)



